const kdf = require('./kdf');
const keccak = require('./keccak');

function splitIntoChunks(data) {
    const chunks = [];
    const chunkSize = 16;
    for (let i = 0; i < data.length; i += chunkSize) {
        const chunk = data.slice(i, i + chunkSize);
        chunks.push(chunk);
    }
    return chunks;
}

class keccakAEAD {
    constructor(iv, key, nonce) {
        const input = kdf.concatUint8Arrays(iv, key, nonce);
        this.state = keccak.SHAKE256(input, 40);
        let r = this.state.slice(0, 16);
        let c = this.state.slice(16, 40);
        let padded_key = kdf.concatUint8Arrays(new Uint8Array(24-key.length), key);
        for (let i = 0; i < padded_key.length; i++) {
            c[i] ^= key[i];
        }
        this.state = kdf.concatUint8Arrays(r, c);
    }

    associated_data_processing(associated_data) {
        const chunks = splitIntoChunks(associated_data);
        let to_xor = null;
        let c = null;
        let r = null;
        let input = null;
        chunks.forEach((chunk) => {
            to_xor = this.state.slice(0, 16);
            c = this.state.slice(16, 40);
            r = new Uint8Array(chunk.length);
            for (let i = 0; i < chunk.length; i++) {
                r[i] = chunk[i] ^ to_xor[i];
            }
            input = kdf.concatUint8Arrays(r, c);
            this.state = keccak.SHAKE256(input, 40);
        });
        r = this.state.slice(0, 16);
        c = this.state.slice(16, 40);
        to_xor = new Uint8Array(24);
        to_xor[23] = 1;
        for (let i = 0; i < c.length; i++) {
            c[i] ^= to_xor[i];
        }
        this.state = kdf.concatUint8Arrays(r, c);
    }

    plaintext_processing(plaintext) {
        const chunks = splitIntoChunks(plaintext);
        let to_xor = null;
        let c = null;
        let r = null;
        let input = null;
        let cipherchunks = [];
        chunks.forEach((chunk) => {
            to_xor = this.state.slice(0, 16);
            c = this.state.slice(16, 40);
            r = new Uint8Array(chunk.length);
            for (let i = 0; i < chunk.length; i++) {
                r[i] = chunk[i] ^ to_xor[i];
            }
            cipherchunks.push(r);
            input = kdf.concatUint8Arrays(r, c);
            this.state = keccak.SHAKE256(input, 40);
        });
        return cipherchunks;
    }

    ciphertext_processing(ciphertext) {
        const chunks = splitIntoChunks(ciphertext);
        let to_xor = null;
        let c = null;
        let r = null;
        let input = null;
        let plaintextchunks = [];
        chunks.forEach((chunk) => {
            to_xor = this.state.slice(0, 16);
            c = this.state.slice(16, 40);
            r = new Uint8Array(chunk.length);
            for (let i = 0; i < chunk.length; i++) {
                r[i] = chunk[i] ^ to_xor[i];
            }
            plaintextchunks.push(r);
            input = kdf.concatUint8Arrays(chunk, c);
            this.state = keccak.SHAKE256(input, 40);
        });
        return plaintextchunks;
    }

    finalize(key) {
        let r = this.state.slice(0, 16);
        let c = this.state.slice(16, 40);
        let padded_key = kdf.concatUint8Arrays(new Uint8Array(24-key.length), key);
        for (let i = 0; i < padded_key.length; i++) {
            c[i] ^= padded_key[i];
        }
        this.state = kdf.concatUint8Arrays(r, c);
        const output = keccak.SHAKE256(this.state, 40);
        let to_xor = output.slice(40-key.length, 40);
        let tag = new Uint8Array(key.length);
        for (let i = 0; i < key.length; i++) {
            tag[i] = key[i] ^ to_xor[i];
        }
        return tag;
    }

    static encrypt(key, plaintext, iv, associated_data, nonce) {
        let sponge = new keccakAEAD(iv, key, nonce);
        sponge.associated_data_processing(associated_data);
        let cipherChunks = sponge.plaintext_processing(plaintext);
        let ciphertext = kdf.concatUint8Arrays(...cipherChunks);
        let tag = sponge.finalize(key);
        return {
            cipher: ciphertext,
            tag: tag
        };
    }

    static decrypt(key, ciphertext, iv, associated_data, nonce) {
        let sponge = new keccakAEAD(iv, key, nonce);
        sponge.associated_data_processing(associated_data);
        let plaintextChunks = sponge.ciphertext_processing(ciphertext);
        let plaintext = kdf.concatUint8Arrays(...plaintextChunks);
        let tag = sponge.finalize(key);
        return {
            plaintext: plaintext,
            tag: tag
        };
    }
}

module.exports = { keccakAEAD };
